MODULE timestep_unstructured_mod
  USE ISO_C_BINDING
  USE OMP_LIB
#ifdef CPP_USING_XIOS
  USE xios
#endif
  USE caldyn_unstructured_mod
  IMPLICIT NONE
  PRIVATE
  SAVE

#define BINDC_(thename) BIND(C, name=#thename)
#define BINDC(thename) BINDC_(dynamico_ ## thename)

#define DBL REAL(C_DOUBLE)
#define DOUBLE1(m) DBL, DIMENSION(m)
#define DOUBLE2(m,n) DBL, DIMENSION(m,n)
#define DOUBLE3(m,n,p) DBL, DIMENSION(m,n,p)
#define DOUBLE4(m,n,p,q) DBL, DIMENSION(m,n,p,q)
#define INDEX INTEGER(C_INT)

#ifdef CPP_USING_XIOS
  TYPE(xios_context) :: ctx_hdl
#endif

CONTAINS

#define FIELD_PS      DOUBLE1(primal_num)
#define FIELD_MASS    DOUBLE2(llm, primal_num)
#define FIELD_Z       DOUBLE2(llm, dual_num)
#define FIELD_U       DOUBLE2(llm, edge_num)
#define FIELD_UL      DOUBLE2(llm+1, edge_num)
#define FIELD_THETA   DOUBLE3(llm, primal_num, nqdyn)
#define FIELD_GEOPOT  DOUBLE2(llm+1, primal_num)

#define HASNAN(field) (ANY(.NOT.ABS(field)<1e20))

  SUBROUTINE caldyn_unstructured(tau, mass_col,rhodz,theta_rhodz,u,geopot,w, & ! IN : flow state
                                 theta,ps,pk,hflux,qv, &    ! OUT : diags (except surface geopot : IN)
                                 dmass_col,drhodz,dtheta_rhodz,du_fast,du_slow, &
                                 dPhi_fast, dPhi_slow, dW_fast, dW_slow) BINDC(caldyn_unstructured) ! OUT : tendencies
    DBL, VALUE :: tau
    FIELD_MASS   :: rhodz, drhodz, pk, berni          ! IN, OUT, DIAG
    FIELD_THETA  :: theta_rhodz, dtheta_rhodz, theta  ! IN, OUT, DIAG
    FIELD_GEOPOT :: wflux, w, geopot, &               ! DIAG, INOUT
         dPhi_fast, dPhi_slow, dW_fast, dW_slow       ! OUT
    FIELD_U      :: u,du_fast,du_slow,hflux,Ftheta,qu ! INOUT,OUT,OUT,DIAG
    FIELD_Z      :: qv                                ! DIAG
    FIELD_PS     :: ps,dmass_col,mass_col             ! OUT,OUT,IN (if eta_mass) or OUT,UNUSED,UNUSED (if eta_lag)
    FIELD_UL     :: wwuu
    DBL          :: time1,time2
    INTEGER :: ij
    
    CALL init_trace()
    !  CALL CPU_TIME(time1)
    time1=OMP_GET_WTIME()
    
    IF(hydrostatic) THEN
       
       !$OMP PARALLEL NUM_THREADS(nb_threads)
       !$OMP DO SCHEDULE(STATIC)
       DO ij=1,edge_num
          du_fast(:,ij)=0.
          du_slow(:,ij)=0.
       END DO
       !$OMP END DO
       CALL compute_theta(mass_col,rhodz,theta_rhodz, theta)
       CALL compute_geopot(rhodz,theta, ps,pk,geopot)
       
       CALL compute_caldyn_fast(tau, pk,berni,theta,geopot, du_fast,u)
       
       CALL compute_pvort_only(rhodz,u,qv,qu)
       CALL compute_caldyn_slow_hydro(rhodz,theta,u, berni,hflux,du_slow)
       CALL compute_coriolis(hflux,theta,qu,Ftheta, drhodz,dtheta_rhodz,du_slow)
       IF(caldyn_eta == eta_mass) THEN
          CALL caldyn_vert(drhodz,rhodz,theta,u, dmass_col,wflux,dtheta_rhodz,du_slow,wwuu)
       END IF
       !$OMP END PARALLEL
       
    ELSE ! NH
       DO ij=1,edge_num
          du_fast(:,ij)=0.
          du_slow(:,ij)=0.
       END DO
       DO ij=1,primal_num
          wflux(1,ij)=0.
          wflux(llm+1,ij)=0.
       END DO
       CALL compute_theta(mass_col,rhodz,theta_rhodz, theta)
       CALL compute_caldyn_solver(tau,rhodz,theta,pk,geopot,W,dPhi_fast,dW_fast,du_fast)
       CALL compute_caldyn_fast(tau, pk,berni,theta,geopot, du_fast,u)
       CALL compute_pvort_only(rhodz,u,qv,qu)
       CALL compute_caldyn_slow_NH(u,rhodz,geopot,W, hflux,du_slow,dPhi_slow,dW_slow) 
       CALL compute_coriolis(hflux,theta,qu,Ftheta, drhodz,dtheta_rhodz,du_slow)
       IF(caldyn_eta == eta_mass) THEN
          CALL caldyn_vert(drhodz,rhodz,theta,u, dmass_col,wflux,dtheta_rhodz,du_slow,wwuu)
          CALL compute_caldyn_vert_NH(rhodz,geopot,W,wflux, du_slow,dPhi_slow,dW_slow)
       END IF
    END IF
    
    time2=OMP_GET_WTIME()
    !  CALL CPU_TIME(time2)
    IF(time2>time1) elapsed = elapsed + time2-time1

    CALL print_trace()
  END SUBROUTINE caldyn_unstructured
  !
  !----------------------------- Time stepping -------------------------------
  
  !
  SUBROUTINE ARK_step(nstep, mass_col,rhodz,theta_rhodz,u,geopot,w, & ! INOUT : flow state
       theta,ps,pk,hflux,qv, &    ! OUT : diags (except surface geopot : IN)
       dmass_col,drhodz,dtheta_rhodz,du_fast,du_slow, &
       dPhi_fast, dPhi_slow, dW_fast, dW_slow) BINDC(ARK_step) ! OUT : tendencies
    INTEGER, VALUE :: nstep ! advance by nstep time steps
    FIELD_PS     :: mass_col, ps                     ! OUT,IN (if eta_mass) or OUT,UNUSED (if eta_lag)
    FIELD_MASS   :: rhodz, pk, berni                 ! IN, DIAG
    FIELD_THETA  :: theta_rhodz, theta               ! IN, DIAG
    FIELD_U      :: u,hflux,qu,Ftheta                ! INOUT,DIAG*3
    FIELD_GEOPOT :: geopot, w, wflux, w_il           ! IN, INOUT, DIAG, DIAG
    FIELD_Z      :: qv                               ! DIAG
    DOUBLE2(       primal_num, max_nb_stage) :: dmass_col    ! OUT
    DOUBLE3(llm,   primal_num, max_nb_stage) :: drhodz   ! OUT
    DOUBLE4(llm,   primal_num, nqdyn, max_nb_stage) :: dtheta_rhodz ! OUT
    DOUBLE3(llm,   edge_num,   max_nb_stage) :: du_fast,du_slow ! OUT
    DOUBLE3(llm+1, primal_num, max_nb_stage) :: &
         dPhi_fast, dPhi_slow, dW_fast, dW_slow      ! OUT
    FIELD_UL     :: DePhil, v_el, G_el, wwuu         ! DIAG*4
    DBL       :: time1,time2
    INTEGER :: step, stage, ij
    
    CALL init_trace()
    !CALL CPU_TIME(time1)
    time1=OMP_GET_WTIME()
    
    !$OMP PARALLEL NUM_THREADS(nb_threads)
    
    DO step=1, nstep
       DO stage=1, nb_stage
          
          IF(hydrostatic) THEN
             
             !$OMP DO SCHEDULE(STATIC)
             DO ij=1,edge_num
                du_fast(:,ij,stage)=0.
                du_slow(:,ij,stage)=0.
             END DO
             
             CALL compute_theta(mass_col,rhodz,theta_rhodz, theta)
             CALL compute_geopot(rhodz,theta, ps,pk,geopot)
             
             CALL compute_caldyn_fast(tauj(stage), pk,berni,theta,geopot, du_fast(:,:,stage), u)
             
             CALL compute_pvort_only(rhodz,u,qv,qu)
             CALL compute_caldyn_slow_hydro(rhodz,theta,u, berni,hflux,du_slow(:,:,stage))
             CALL compute_coriolis(hflux,theta,qu,Ftheta, drhodz(:,:,stage), dtheta_rhodz(:,:,:,stage),du_slow(:,:,stage))
             IF(caldyn_eta == eta_mass) THEN
                CALL caldyn_vert(drhodz(:,:,stage),rhodz,theta,u, &
                     dmass_col(:,stage),wflux,dtheta_rhodz(:,:,:,stage),du_slow(:,:,stage),wwuu)
             END IF
             
          ELSE ! NH
             !$OMP DO SCHEDULE(STATIC)
             DO ij=1,primal_num
                wflux(1,ij)=0.
                wflux(llm+1,ij)=0.
             END DO
             !$OMP END DO
             CALL compute_theta(mass_col,rhodz,theta_rhodz, theta)
             CALL compute_caldyn_solver(tauj(stage),rhodz,theta,pk,geopot,W, &
                  dPhi_fast(:,:,stage), dW_fast(:,:,stage), du_fast(:,:,stage))
             CALL compute_caldyn_fast(tauj(stage), pk,berni,theta,geopot, du_fast(:,:,stage),u)
             CALL compute_pvort_only(rhodz,u,qv,qu)
             CALL compute_caldyn_slow_NH(u,rhodz,geopot,W, hflux, du_slow(:,:,stage), dPhi_slow(:,:,stage), dW_slow(:,:,stage))
             CALL compute_coriolis(hflux,theta,qu,Ftheta, drhodz(:,:,stage),dtheta_rhodz(:,:,:,stage), du_slow(:,:,stage))
             IF(caldyn_eta == eta_mass) THEN
                CALL caldyn_vert(drhodz(:,:,stage),rhodz,theta,u, &
                     dmass_col(:,stage),wflux,dtheta_rhodz(:,:,:,stage), du_slow(:,:,stage), wwuu)
                CALL compute_caldyn_vert_NH(rhodz,geopot,W,wflux, du_slow(:,:,stage), dPhi_slow(:,:,stage), dW_slow(:,:,stage) )
             END IF
          END IF ! NH
          
          ! FIXME : mass_col is computed from rhodz
          ! so that the DOFs are the same whatever caldyn_eta
          ! in DYNAMICO mass_col is prognosed rather than rhodz
          
#define UPDATE(clj,field,dfield) update(stage,SIZE(field),clj,field,dfield)
          
          CALL UPDATE(cslj, rhodz, drhodz)
          CALL UPDATE(cslj, theta_rhodz, dtheta_rhodz)
          CALL UPDATE(cslj, u, du_slow)
          CALL UPDATE(cflj, u, du_fast)
          
          IF(.NOT.hydrostatic) THEN
             CALL UPDATE(cslj, geopot, dPhi_slow)
             CALL UPDATE(cflj, geopot, dPhi_fast)
             CALL UPDATE(cslj, W, dW_slow)
             CALL UPDATE(cflj, W, dW_fast)
          END IF
          
       END DO
    END DO
    !$OMP END PARALLEL
    
    time2=OMP_GET_WTIME()
    !  CALL CPU_TIME(time2)
    IF(time2>time1) elapsed = elapsed + time2-time1

    CALL print_trace()
    
  END SUBROUTINE ARK_step
  
  
  SUBROUTINE update(j,sz,clj,field,dfield)
    INTEGER :: j, sz ! stage in ARK scheme, field size
    DOUBLE2(max_nb_stage,max_nb_stage) :: clj ! modified Butcher tableau
    DOUBLE1(sz) :: field
    DOUBLE2(sz, max_nb_stage) :: dfield
    !
    INTEGER :: l, ij
    CALL enter_trace(id_update, 8*sz*(j+1) )
    !  PRINT *, clj(1:j,j)
    SELECT CASE(j)
    CASE(1)
       !$OMP DO SCHEDULE(static)
       DO ij=1,sz
          field(ij) = field(ij) &
               + clj(1,j)*dfield(ij,1)
       END DO
    CASE(2)
       !$OMP DO SCHEDULE(static)
       DO ij=1,sz
          field(ij) = field(ij) &
               + clj(1,j)*dfield(ij,1) &
               + clj(2,j)*dfield(ij,2)
       END DO
    CASE(3)
       !$OMP DO SCHEDULE(static)
       DO ij=1,sz
          field(ij) = field(ij) &
               + clj(1,j)*dfield(ij,1) &
               + clj(2,j)*dfield(ij,2) &
               + clj(3,j)*dfield(ij,3)
          
       END DO
    CASE(4)
       !$OMP DO SCHEDULE(static)
       DO ij=1,sz
          field(ij) = field(ij) &
               + clj(1,j)*dfield(ij,1) &
               + clj(2,j)*dfield(ij,2) &
               + clj(3,j)*dfield(ij,3) &
               + clj(4,j)*dfield(ij,4)
       END DO
    CASE(5)
       !$OMP DO SCHEDULE(static)
       DO ij=1,sz
          field(ij) = field(ij) &
               + clj(1,j)*dfield(ij,1) &
               + clj(2,j)*dfield(ij,2) &
               + clj(3,j)*dfield(ij,3) &
               + clj(4,j)*dfield(ij,4) &
               + clj(5,j)*dfield(ij,5)
       END DO
    END SELECT
    CALL exit_trace()
  END SUBROUTINE update
  
  !---------------------------------------------- XIOS ----------------------------------------

#ifdef CPP_USING_XIOS
  
  SUBROUTINE setup_xios() BINDC(setup_xios)
    ! MPI_INIT / MPI_finalize are assumed to be called BEFORE/AFTER this routine
    ! This is the case when calling from Python after importing mpi4py
    INTEGER :: ierr, mpi_threading_mode
    
    PRINT *, 'Initialize XIOS and obtain our communicator'
    CALL xios_initialize("icosagcm",return_comm=comm_icosa)
    
    PRINT *, 'Initialize our XIOS context'
    
    CALL xios_context_initialize("icosagcm",comm_icosa)
    CALL xios_get_handle("icosagcm",ctx_hdl)
    CALL xios_set_current_context(ctx_hdl)
    !   CALL xios_set_axis_attr("lev",n_glo=llm ,value=lev_value) ;
    !   CALL xios_set_axis_attr("levp1",n_glo=llm+1 ,value=lev_valuep1) ;
    !   CALL xios_set_axis_attr("nq",n_glo=nqtot, value=nq_value) ;
    !   CALL xios_set_domaingroup_attr("i",ni_glo=ncell_tot, ibegin=displ, ni=ncell)
    !   CALL xios_set_domaingroup_attr("i", data_dim=1, type='unstructured' , nvertex=6, i_index=ind_glo)
    !   CALL xios_set_domaingroup_attr("i",lonvalue_1d=lon, latvalue_1d=lat, bounds_lon_1d=bounds_lon, bounds_lat_1d=bounds_lat)
    !   CALL xios_set_domain_attr("u",ni_glo=ncell_tot, ibegin=displ, ni=ncell)
    !   CALL xios_set_domain_attr("u", data_dim=1, type='unstructured' , nvertex=2, i_index=ind_glo)
    !  CALL xios_set_domain_attr("u",lonvalue_1d=lon, latvalue_1d=lat, bounds_lon_1d=bounds_lon, bounds_lat_1d=bounds_lat)
    !   CALL xios_set_domain_attr("v",ni_glo=ncell_tot, ibegin=displ, ni=ncell)
    !   CALL xios_set_domain_attr("v", data_dim=1, type='unstructured' , nvertex=3)
    !   CALL xios_set_domain_attr("v",lonvalue_1d=lon, latvalue_1d=lat, bounds_lon_1d=bounds_lon, bounds_lat_1d=bounds_lat)
    !   CALL xios_set_timestep(dtime)  
    !   CALL xios_set_fieldgroup_attr("standard_output", freq_op=itau_out*xios_timestep, freq_offset=(itau_out-1)*xios_timestep)    
    ! CALL xios_close_context_definition()
    
    ! PRINT *, 'Read data'
    ! CALL xios_recv_field(name,field)
    
    ! PRINT *,'Write data'
    ! CALL xios_send_field(name,field_tmp)
    
   !   PRINT *, 'Flush to disk, clean up and die'
   !   CALL xios_context_finalize
  END SUBROUTINE setup_xios
!
  SUBROUTINE call_xios_set_timestep(dt) BINDC(xios_set_timestep)
    DBL, VALUE :: dt
    TYPE(xios_duration)      :: dtime
    dtime%second=dt
    CALL xios_set_timestep(dtime)
  END SUBROUTINE call_xios_set_timestep

  SUBROUTINE call_xios_update_calendar(step) BINDC(xios_update_calendar)
    INDEX, value :: step
    CALL xios_update_calendar(step)
  END SUBROUTINE call_xios_update_calendar

#endif

END MODULE timestep_unstructured_mod
